from typing import Any

import google.generativeai as genai

from .agent import Agent

class GeminiAgent(Agent):
    model: str
    context: str
    message: str
    def __init__(self, api_key: str, model: str = "models/gemini-pro", **kwargs) -> None:
        super().__init__()
        self.api_key = api_key
        self.model = model
        self.message = ""
        genai.configure(api_key=api_key)
        self.model = genai.GenerativeModel(model_name=model)
        self.chat = self.model.start_chat()

    def receive_prompt(self, prompt: str) -> Any:
        self.message = prompt

    def _postprocess(self, response: str) -> str:
        return response.strip()

    def generate_reply(self) -> Any:
        response = self.chat.send_message(self.message)
        return self._postprocess(response.text)



import json
import textwrap
from typing import Any, List

import numpy as np

from .game import Game


class BeautyContest(Game):
    upper_bound: int
    prize: int
    chosen_numbers: List[float]
    fouls: List[bool]
    status: bool
    winners: List[int]
    payoffs: List[float]

    def _set_rule(self) -> None:
        """Set the rules for the Beauty Contest game.

        Configures prompt and key for extraction.
        """
        self.rule_prompt = textwrap.dedent("\
            You are playing a game.\n\
            The basic rules are as follows:\n\
                - it consists of {} players, including you.\n\
                - this is a one-round game.\n\
                - you need to choose a real number between 0 and {}, 0 and {} inclusive.\n\
                - everyone does not know how others would make choices beforehand.\n\
                - the winner is the person whose number is the closest to the two thirds of the average of all chosen numbers.\n\
                - the winner gets a fixed prize, in case of a tie the prize is split amongst those who tie.\n\
                - your goal is primarily to maximize the possibility of getting the prize and secondly to maximize the your prize.\n\
            \n\
            Please just strictly output a JSON string, which has following keys:\n\
                - understanding: str, your brief understanding of the game\n\
                - choice: float, the number which you would like to choose\n\
                - reason: str, the brief reason why you made the choice\n\
        ")
        self.key_to_extract = "choice"

    def __init__(self, num_players: int, upper_bound: int = 100, prize: int = 1) -> None:
        """
        Initialize the BeautyContest instance.

        Args:
            num_players (int): The number of players in the game.
            upper_bound (int, optional): The upper bound for choosing numbers.
            prize (int, optional): The fixed prize for the winner.
        """
        super().__init__(num_players)
        self.upper_bound = upper_bound
        self.prize = prize
        self.chosen_numbers = [np.NaN] * self.num_players
        self._set_rule()

    def get_rule(self, idx: int) -> str:
        """
        Get the rule prompt for a specific player index after formatting.

        Insert the number of players and the upper bound for choosing numbers
        into the rule prompt string and return.

        Args:
            idx (int): The index of the player.

        Returns:
            str: The rule prompt after formatting.
        """
        return self.rule_prompt.format(self.num_players, self.upper_bound, self.upper_bound)

    def extract_from_raw_reply(self, reply: str):
        """
        Extract information from the raw reply string and return the choosing score.

        Args:
            reply (str): The raw reply string.

        Returns:
            float: The choosing score.
        """
        try:
            pos_fr, pos_to = reply.find("{"), reply.rfind("}")
            json_content = reply[pos_fr: pos_to + 1]
            reply_d = json.loads(json_content)
            reply_key = float(reply_d[self.key_to_extract])
        except Exception:
            return np.NaN
        return reply_key

    def make_decision(self, idx: int, reply: str) -> None:
        """
        Make a decision for a player based on their reply.

        Args:
            idx (int): The index of the player.
            reply (str): The player's reply.
        """
        self.chosen_numbers[idx] = self.extract_from_raw_reply(reply)

    def feedback_result(self, **kwarg) -> Any:
        """
        Provide feedback on the game result.

        Returns:
            dict: A dictionary containing relevant information.For
            example:

            {'upper_bound': 100,
             'choices': [0.0, 33.33, 33.33],
             'status': True,
             'fouls': [False, False, False],
             'winners': [0],
             'payoffs': [1.0, 0.0, 0.0]}
        """
        self.fouls = [False if not np.isnan(value) and 0 <= value <= self.upper_bound else True
                      for value in self.chosen_numbers]

        count_valid = sum([1 if not br else 0
                           for br in self.fouls])
        self.status = True if count_valid >= 1 else False

        avg = np.nanmean(self.chosen_numbers)
        ctr = avg * 2 / 3
        dif = [abs(val - ctr) if not self.fouls[idx] else np.NaN
               for (idx, val) in enumerate(self.chosen_numbers)]
        self.winners = list(np.where(dif == np.nanmin(dif))[0])

        self.payoffs = [0.0 if not self.fouls[idx] else np.NaN
                        for idx in range(self.num_players)]
        for winner in self.winners:
            self.payoffs[winner] = 1.0 * self.prize / len(self.winners)

        result = {
            "upper_bound": self.upper_bound,
            "choices": self.chosen_numbers,
            "status": self.status,
            "fouls": self.fouls,
            "winners": self.winners,
            "payoffs": self.payoffs
        }
        return result



from abc import ABC, abstractmethod
import random
from typing import Any, List

import toml

from ..games import *
from ..players import *

class Host(ABC):
    config: dict
    apis: dict
    game: Game
    num_players: int
    players: List[Player]
    def __init__(self, config: dict) -> None:
        self.config = config
        self.apis = None
    
    def set_apis(self, apis: dict) -> None:
        self.apis = apis
    
    def initialize_game(self) -> Any:
        assert self.apis is not None, "api information for LLM agents is not set"
        num_players = self.config["num_players"]
        self.num_players = int(num_players)
        players_class = self.config["players_class"]
        self.players = [eval(player_class[0])(
                            api_base=self.apis[player_class[0]].get("api_base"),
                            api_key=self.apis[player_class[0]].get("api_key"),
                            secret_key=self.apis[player_class[0]].get("secret_key"),
                            model=player_class[1]
                        ) for player_class in players_class]
        game_class = self.config["game_class"]
        self.game = eval(game_class)(
                        num_players=self.num_players
                    )

    def run_game(self) -> Any:
        self.initialize_game()
        for index in range(self.num_players):
            rule_prompt = self.game.get_rule(idx=index)
            self.players[index].receive_prompt(rule_prompt)
            reply = self.players[index].generate_reply()
            # print(f"[Agent {index} - {self.players[index]}]")
            # print(reply)
            self.game.make_decision(idx=index, reply=reply)
        result = self.game.feedback_result()
        return result


from typing import Any

import google.generativeai as palm

from .agent import Agent

class PaLMAgent(Agent):
    model: str
    context: str
    messages: str
    def __init__(self, api_key: str, model: str = "models/chat-bison-001", **kwargs) -> None:
        super().__init__()
        self.api_key = api_key
        self.model = model
        self.context = ""
        self.messages = ""
        palm.configure(api_key=api_key)

    def receive_prompt(self, prompt: str) -> Any:
        self.messages = prompt

    def _postprocess(self, response: str) -> str:
        return response.strip()

    def generate_reply(self) -> Any:
        response = palm.chat(
            context=self.context,
            messages=self.messages,
            model=self.model
        )
        return self._postprocess(response.last)



from abc import abstractmethod
from typing import Any

from .player import Player

class Agent(Player):
    api_base: str
    api_key: str
    organization: str
    def __init__(self, **kwargs) -> None:
        super().__init__()
        self.api_base = None
        self.api_key = None
        self.organization = None

    @abstractmethod
    def receive_prompt(self, prompt: str) -> Any:
        raise NotImplementedError

    @abstractmethod
    def generate_reply(self) -> Any:
        raise NotImplementedError


from abc import ABC, abstractmethod
from typing import Any

class Player(ABC):
    def __init__(self) -> None:
        pass
    
    @abstractmethod
    def receive_prompt(self, prompt: str) -> Any:
        raise NotImplementedError

    @abstractmethod
    def generate_reply(self) -> Any:
        raise NotImplementedError


import hashlib
import json
import time
from typing import Any, Optional

import requests

from .agent import Agent

def calculate_md5(input_string):
    md5 = hashlib.md5()
    md5.update(input_string.encode("utf-8"))
    encrypted = md5.hexdigest()
    return encrypted

class BaichuanAgent(Agent):
    secret_key: str
    model: str
    def __init__(self, api_base: Optional[str] = None, api_key: str = None, secret_key: Optional[str] = None, model: str = "Baichuan2", **kwargs) -> None:
        super().__init__()
        self.api_base = api_base
        self.api_key = api_key
        self.secret_key = secret_key
        self.model = model
        self.history = []

    def _preprocess(self, role: str, content: str) -> dict:
        return {"role": role, "content": content}

    def receive_prompt(self, prompt: str) -> Any:
        self.history.append(self._preprocess(role="user", content=prompt))

    def _postprocess(self, response: dict) -> str:
        role = response["data"]["messages"][0]["role"] or "assistant"
        content = response["data"]["messages"][0]["content"]
        self.history.append(self._preprocess(role=role, content=content))
        return content

    def generate_reply(self) -> Any:
        data = {
            "model": self.model,
            "messages": self.history
        }
        json_data = json.dumps(data)
        time_stamp = int(time.time())
        signature = calculate_md5(self.secret_key + json_data + str(time_stamp))
        headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self.api_key}",
            "X-BC-Request-Id": "your requestId",
            "X-BC-Timestamp": str(time_stamp),
            "X-BC-Signature": signature,
            "X-BC-Sign-Algo": "MD5",
        }
        response = requests.post(self.api_base, data=json_data, headers=headers)
        return self._postprocess(eval(response.text))



import json
import textwrap
from typing import Any, List

import numpy as np

from .game import Game


class SecondPriceAuction(Game):
    budget: int
    private_values: List[int]
    bidding_prices: List[float]
    fouls: List[bool]
    status: bool
    winner: int
    payoffs: List[float]

    def _set_rule(self) -> None:
        """Set the rules for the Second Price Auction game.

        Configures prompt and key for extraction.
        """
        self.rule_prompt = textwrap.dedent("\
            You are participating in an auction.\n\
            The basic rules are as follows:\n\
                - it consists of {} bidders, included you.\n\
                - this is a one-round auction.\n\
                - there is only 1 item, and your private value of the item is {} units(private values may vary among bidders).\n\
                - you have {} units of assets, and you need to place a bid which is not higher than your assets.\n\
                - everyone does not know either private value of others or how others would make choices beforehand.\n\
                - The bidder who places the highest bid among all the bids will get the item, and only need to pay an amount of assets equalling to the second-highest bid among all the bids.\n\
                - If there are multiple highest bid, only the bidder with minimal index by order will get the item.\n\
                - If you get the item, your payoff equals to your remaining assets(=assets deducting payment) plus your private value, otherwise your payoff equals to your original assets.\n\
                - your goal is to maximize your overall payoffs(notice that getting the item is not necessary).\n\
            \n\
            Please just strictly output a JSON string, which has following keys:\n\
                - understanding: str, your brief understanding of the auction\n\
                - choice: float, the price which you would like to place the bid at\n\
                - reason: str, the brief reason why you made the choice\n\
        ")
        self.key_to_extract = "choice"

    def __init__(self, num_players: int, budget: int = 1000, mean_private_value: int = 500,
                 std_private_value: int = 100) -> None:
        """
        Initialize the Second Price Auction instance.

        Draw random sample from a normal distribution
        and assign it to the private value of specific player.
        Args:
            num_players (int): The number of players in the game.
            budget (int, optional): The budget for each player.
            mean_private_value (int, optional): The mean value for private values.
            std_private_value (int, optional): The standard deviation for private values.
        """

        super().__init__(num_players)
        self.budget = budget
        self.private_values = []
        while len(self.private_values) < self.num_players:
            value = int(np.random.normal(mean_private_value, std_private_value))
            if 0 < value <= self.budget:
                self.private_values.append(value)
        self.bidding_prices = [np.NaN] * self.num_players
        self._set_rule()

    def get_rule(self, idx: int) -> str:
        """
        Get the rule prompt for a specific player.

        Insert the number of players, the private values of the player and
        the upper bound for choosing numbers into the rule prompt string and return.

        Args:
            idx (int): The index of the player.

        Returns:
            str: The rule prompt.
        """

        return self.rule_prompt.format(self.num_players, self.private_values[idx], self.budget)

    def extract_from_raw_reply(self, reply: str):
        """
        Extract information from the raw reply string and return the bidding price.

        Args:
            reply (str): The raw reply string.

        Returns:
            float: The bidding price.
        """
        try:
            pos_fr, pos_to = reply.find("{"), reply.rfind("}")
            json_content = reply[pos_fr: pos_to + 1]
            reply_d = json.loads(json_content)
            reply_key = float(reply_d[self.key_to_extract])
        except Exception:
            return np.NaN
        return reply_key

    def make_decision(self, idx: int, reply: str) -> None:
        """
        Make a decision for a player based on their reply.

        Args:
            idx (int): The index of the player.
            reply (str): The player's reply.
        """
        self.bidding_prices[idx] = self.extract_from_raw_reply(reply)

    def feedback_result(self, **kwarg) -> Any:
        """
        Provide feedback on the game result.

        Returns:
            dict: A dictionary containing relevant information.For
            example:

            {'budget': 1000,
            'private_values': [657, 503, 327],
            'choices': [657.0, 502.0, 600.0],
            'status': True,
            'fouls': [False, False, False],
            'winner': 0,
            'payoffs': [1057.0, 1000.0, 1000.0]}
        """
        self.fouls = [False if not np.isnan(value) and 0 <= value <= self.budget else True
                      for value in self.bidding_prices]

        count_valid = sum([1 if not br else 0
                           for br in self.fouls])
        self.status = True if count_valid >= 2 else False

        self.winner = np.nanargmax(self.bidding_prices)

        self.payoffs = [1.0 * self.budget if not self.fouls[idx] else np.NaN
                        for idx in range(self.num_players)]
        profit = (self.private_values[self.winner] - np.nanmax(np.delete(self.bidding_prices, self.winner))
                  if self.status else np.NaN
                  )
        self.payoffs[self.winner] = self.budget + profit

        result = {
            "budget": self.budget,
            "private_values": self.private_values,
            "choices": self.bidding_prices,
            "status": self.status,
            "fouls": self.fouls,
            "winner": self.winner,
            "payoffs": self.payoffs
        }
        return result



from abc import ABC, abstractmethod
from typing import Any


class Game(ABC):
    num_players: int
    rule_prompt: str
    key_to_extract: str

    def __init__(self, num_players: int, **kwargs) -> Any:
        self.num_players = num_players
        self.rule_prompt = None
        self.key_to_extract = None

    @abstractmethod
    def get_rule(self, **kwargs) -> Any:
        raise NotImplementedError

    @abstractmethod
    def make_decision(self, **kwargs) -> Any:
        raise NotImplementedError

    @abstractmethod
    def feedback_result(self, **kwarg) -> Any:
        raise NotImplementedError



from typing import Any, Optional

import openai

from .agent import Agent

class GPTAgent(Agent):
    model: str
    history: list
    def __init__(self, api_base: Optional[str] = None, api_key: str = None, model: str = "gpt-3.5-turbo", **kwargs) -> None:
        super().__init__()
        self.api_base = api_base
        self.api_key = api_key
        self.model = model
        self.history = []
        openai.api_base = api_base
        openai.api_key = api_key
    
    def _preprocess(self, role: str, content: str) -> dict:
        return {"role": role, "content": content}

    def receive_prompt(self, prompt: str) -> Any:
        self.history.append(self._preprocess(role="user", content=prompt))

    def _postprocess(self, response: dict) -> str:
        role = response["choices"][0]["message"]["role"] or "assistant"
        content = response["choices"][0]["message"]["content"].strip()
        self.history.append(self._preprocess(role=role, content=content))
        return content

    def generate_reply(self) -> Any:
        response = openai.ChatCompletion.create(
            model=self.model,
            messages=self.history
        )
        return self._postprocess(response)



from .player import Player
from .agent import Agent
from .baichuan_agent import BaichuanAgent
from .gemini_agent import GeminiAgent
from .gpt_agent import GPTAgent
from .palm_agent import PaLMAgent

__all__ = ["Player", "Agent", "BaichuanAgent", "GeminiAgent", "GPTAgent", "PaLMAgent"]